/**
 * 各Webブラウザによる、イベント実装の違いを吸収するクラスです。
 * DOM Events Specificationをベースに実装しています。
 * 但し、全てを実装してるわけではありません。
 *
 * Version 0.9.0.0
 *
 * このプログラムのライセンスについて。
 * このプログラムは、MITライセンス(The MIT License)の元で公開します。
 * Released under the MIT Licenses.
 *
 * ライセンスの原文は、以下のURLを参照してください。
 * http://www.yscjp.com/licenses/MIT_LICENSE.txt
 *
 * Copyright (c) 2010-2012 Yagyu Teru, http://yscjp.com/
 *
 */

var com;
if (!com) com = {};
if (!com.yscjp) com.yscjp = {};
if (!com.yscjp.jsapp) com.yscjp.jsapp = {};

/*
 * 名前空間の作成
 */
com.yscjp.jsapp.JSEvent = {};

(function() {

    var JSEvent = com.yscjp.jsapp.JSEvent;
    var listenerTbl = new Array();

    /**
     * 各Webブラウザによる、イベント実装の違いを吸収するための、ユーティリティ
     */
    function createEvent(e) {

        e = getEvent(e);

        if (e instanceof com.yscjp.jsapp.JSEventBace) {

            return e;
        }

        switch (e.type) {
            case "load"   :
            case "unload" :
            case "abort"  :
            case "error"  :
            case "select" :
            case "change" :
            case "submit" :
            case "reset"  :
            case "resize" :
            case "scroll" :
            case "focus"  :
            case "blur"   :
                return new com.yscjp.jsapp.JSEventBace(e);
            case "keypress" :
            case "keydown"  :
            case "keyup"    :
                return new com.yscjp.jsapp.JSKeyboardEvent(e);
            case "click"     :
            case "dblclick"  :
            case "mousedown" :
            case "mouseup"   :
            case "mouseover" :
            case "mousemove" :
            case "mouseout"  :
                return new com.yscjp.jsapp.JSMouseEvent(e);
            case "textInput" :
                return new com.yscjp.jsapp.JSTextEvent(e);
            default :
                return new com.yscjp.jsapp.JSEventBace(e);
                //return e;
        }
    };


    /**
     * ネイティブイベントの取得
     */
    function getEvent(evt) {

        if (evt) {
        } else if (window.event) {

            evt = event;

        } else {

            /*
             * イベントが存在しない。
             */
            throw new Error("event not supported.");
        }

        return evt;
    };


    /**
     * イベントリスナの登録
     */
    function addEventListener(eventTarget, type, listener, useCapture) {

        var listenerValue = new Object();
        listenerValue.target = eventTarget;
        listenerValue.type = type;
        listenerValue.listener = listener;

        if (eventTarget.addEventListener) {

            listenerValue.func = function (e) {

                var evt = createEvent(e);
                listener(evt);
            };
            listenerTbl.push(listenerValue);

            /*
             * W3Cの標準インターフェイスが実装されている環境であれば、
             * それを呼び出すだけ
             */
            eventTarget.addEventListener(type, listenerValue.func, useCapture);

        } else if (eventTarget.attachEvent) {

            var oldListenerValue = getListenerValue(eventTarget, type, listener, true);

            listenerValue.func = function () {

                var ownerDocument = eventTarget.ownerDocument;
                if (null == ownerDocument) {
                    ownerDocument = eventTarget;
                }
                var parentWindow = ownerDocument.parentWindow;
                if (null == parentWindow) {
                    parentWindow = eventTarget;
                }

                var e = parentWindow.event;
                e.target = e.srcElement;
                e.currentTarget = eventTarget;
                e = createEvent(e);
                listener(e);
            };
            listenerTbl.push(listenerValue);

            /*
             * IE特有のイベント登録インターフェイス
             * 同じターゲットに同じイベントタイプで同じハンドラを登録すると、
             * IEの場合登録したハンドラが全て有効となってしまうので、
             * 既に存在すれば、一旦削除してから登録する。
             */
            if (oldListenerValue) {

                eventTarget.detachEvent("on" + type, oldListenerValue.func);
            }
            eventTarget.attachEvent("on" + type, listenerValue.func);

        } else {

            /*
             * 簡単なリスナの登録
             * この方法では、ひとつの要素の同一イベントに
             * 複数のリスナを登録できない。
             */
            eventTarget["on" + type] = listener;
        }
    };


    /**
     * イベントリスナの解放
     */
    function removeEventListener(eventTarget, type, listener, useCapture) {

        var func = null;
        var listenerValue = getListenerValue(eventTarget, type, listener, true);
        if (null != listenerValue) {

            func = listenerValue.func;

        } else {

            func = listener;
        }

        if (eventTarget.removeEventListener) {

            /*
             * W3Cの標準インターフェイスが実装されている環境であれば、
             * それを呼び出すだけ
             */
            eventTarget.removeEventListener(type, func, useCapture);

        } else if (eventTarget.detachEvent) {

            /*
             * IE特有のイベント登録インターフェイス
             */
            eventTarget.detachEvent("on" + type, func);

        } else {

            /*
             * 簡単なリスナの登録
             * この方法では、ひとつの要素の同一イベントに
             * 複数のリスナを登録できない。
             */
            eventTarget["on" + type] = null;
        }
    };


    function getListenerValue(eventTarget, type, listener, isRemove) {

        if ("undefined" == isRemove) {

            isRemove = false;
        }

        for (var i = 0; i < listenerTbl.length; i++) {

            var listenerValue = listenerTbl[i];
            if (listenerValue.target == eventTarget &&
                listenerValue.type == type &&
                listenerValue.listener == listener) {

                if (isRemove) {

                    listenerTbl.splice(i, 1);

                }

                return listenerValue;
            }
        }

        return null;
    };


    JSEvent.createEvent = createEvent;
    JSEvent.getEvent = getEvent;
    JSEvent.addEventListener = addEventListener;
    JSEvent.removeEventListener = removeEventListener;

})();

/******************************************************************************
 * イベントベースクラス
 * DOM Events Specificationの、Eventインターフェイスに該当する。
 *****************************************************************************/
com.yscjp.jsapp.JSEventBace = function (e) {

//    this.currentTarget;  // イベントハンドラが設定されている要素
//    this.target;         // 実際にイベント発行の元となった要素
//
//    /* Webブラウザから取得した、環境固有のイベント */
//    this.nativeEvent = null;
//    /* イベントのタイプ（名称） */
//    this.type = null;
//    /* 実際にイベント発行の元となった要素 */
//    this.target = null;
//    /* イベントハンドラが設定されている要素 */
//    this.currentTarget = null;

    if (e) {

        this.initJSEventBace(e);
    }
};


/**
 * インスタンスの初期化
 * コンストラクタに引数を指定できないので、オブジェクトのインスタンスを
 * 外部からの値で初期化する必要がある場合は、このようにする。
 */
com.yscjp.jsapp.JSEventBace.prototype.initJSEventBace = function (e) {

    var JSEvent = com.yscjp.jsapp.JSEvent;

    this.nativeEvent = JSEvent.getEvent(e);

    if ("undefined" != typeof this.nativeEvent.type) {

        this.type = this.nativeEvent.type;

    } else {
    }

    if ("undefined" != typeof this.nativeEvent.currentTarget) {

        /*
         * currentTargetが定義されていれば、targetも定義されている（はず）
         */
        this.target = this.nativeEvent.target;
        this.currentTarget = this.nativeEvent.currentTarget;

    } else if ("undefined" != typeof this.nativeEvent.srcElement) {

        this.target = this.nativeEvent.srcElement;
        this.currentTarget = this.target;

    } else {
    }
};


/**
 * イベント伝播の防止
 * イベントが、上位の要素へ伝播するのを防止する。
 */
com.yscjp.jsapp.JSEventBace.prototype.stopPropagation = function () {

    if (this.nativeEvent.stopPropagation) {

        /*
         * W3Cの標準インターフェイスが実装されている環境であれば、
         * それを呼び出すだけ
         */
        this.nativeEvent.stopPropagation();

    } else if ("undefined" != typeof this.nativeEvent.cancelBubble) {

        /*
         * IE4以上でサポートされたプロパティを使う
         */
        this.nativeEvent.cancelBubble = true;

    } else {

        /*
         * イベント伝播の防止はできないので、例外の送出する。
         */
        throw new Error("stopPropagation not supported.");
    }
};


/**
 * デフォルト動作の防止
 * イベントによる、デフォルトの動作を防止する。
 */
com.yscjp.jsapp.JSEventBace.prototype.preventDefault = function () {

    if (this.nativeEvent.preventDefault) {

        /*
         * W3Cの標準インターフェイスが実装されている環境であれば、
         * それを呼び出すだけ
         */
        this.nativeEvent.preventDefault();

    } else {

        /*
         * IE4以上でサポートされたプロパティを使う
         * 万が一、実装していないWebブラウザの場合も、どのみち無効なので、
         * それでよしとする。
         * シビアなシステムには、この機能は使ってはいけない。
         * Alt+hなどで、メニューを表示するような処理は、無効にはできない。
         */
        this.nativeEvent.returnValue = false;
    }
};


/******************************************************************************
 * テキストイベント
 * DOM Events Specificationの、TextEventインターフェイスに該当する。
 *****************************************************************************/
com.yscjp.jsapp.JSTextEvent = function (e) {

//    this.data = null;

    if (e) {

        this.initJSTextEvent(e);
    }
};
com.yscjp.jsapp.JSTextEvent.prototype = new com.yscjp.jsapp.JSEventBace();


/**
 * インスタンスの初期化
 * コンストラクタに引数を指定できないので、オブジェクトのインスタンスを
 * 外部からの値で初期化する必要がある場合は、このようにする。
 */
com.yscjp.jsapp.JSTextEvent.prototype.initJSTextEvent = function (e) {

    /*
     * 必ず継承元の初期化メソッドを、呼び出す必要がある。
     */
    this.initJSEventBace(e);
};


/******************************************************************************
 * インプットイベント
 * DOM Events Specificationに該当するインターフェイスは存在しない。
 *****************************************************************************/
com.yscjp.jsapp.JSInputEvent = function(e) {

//    /* コントロールキーが押されていたか */
//    this.ctrlKey = null;
//    /* シフトキーが押されていたか */
//    this.shiftKey = null;
//    /* ALTキーが押されていたか */
//    this.altKey = null;
//    /* METAキーが押されていたか */
//    this.metaKey = null;

    if (e) {

        this.initJSInputEvent(e);
    }
};
com.yscjp.jsapp.JSInputEvent.prototype = new com.yscjp.jsapp.JSEventBace();


/**
 * インスタンスの初期化
 * コンストラクタに引数を指定できないので、オブジェクトのインスタンスを
 * 外部からの値で初期化する必要がある場合は、このようにする。
 */
com.yscjp.jsapp.JSInputEvent.prototype.initJSInputEvent = function (e) {

    /*
     * 必ず継承元の初期化メソッドを、呼び出す必要がある。
     */
    this.initJSEventBace(e);

    if ("undefined" != typeof this.nativeEvent.ctrlKey) {

        this.ctrlKey = this.nativeEvent.ctrlKey;
    }

    if ("undefined" != typeof this.nativeEvent.shiftKey) {

        this.shiftKey = this.nativeEvent.shiftKey;
    }

    if ("undefined" != typeof this.nativeEvent.altKey) {

        this.altKey = this.nativeEvent.altKey;

    } else {

        if ("undefined" != typeof this.nativeEvent.metaKey) {

            /*
             * altKeyとmetaKeyは、同じ意味で使用されることが多いので、
             * 個別に定義されていない場合は、同じものを設定する。
             */
            this.altKey = this.nativeEvent.metaKey;
        }
    }

    if ("undefined" != typeof this.nativeEvent.metaKey) {

        this.metaKey = this.nativeEvent.metaKey;

    } else {

        /*
         * altKeyとmetaKeyは、同じ意味で使用されることが多いので、
         * 個別に定義されていない場合は、同じものを設定する。
         */
        this.metaKey = this.altKey;
    }
};


/******************************************************************************
 * マウスイベント
 * DOM Events Specificationの、MouseEventインターフェイスに該当する。
 *****************************************************************************/
com.yscjp.jsapp.JSMouseEvent = function(e) {

//    /* カーソルのスクリーン上での相対X座標 */
//    this.screenX = null;
//    /* カーソルのスクリーン上での相対Y座標 */
//    this.screenY = null;
//    /* カーソルのページ上での相対X座標 */
//    this.clientX = null;
//    /* カーソルのページ上での相対Y座標 */
//    this.clientY = null;
//    /* 押されたマウスボタンを表す整数 */
//    this.button = null;
//    /*  */
//    this.relatedTarget = null;

    /*
     * 初期化メソッドを呼び出す。
     */
    if (e) {

        this.initJSMouseEvent(e);
    }
};
com.yscjp.jsapp.JSMouseEvent.prototype = new com.yscjp.jsapp.JSInputEvent();


/**
 * インスタンスの初期化
 * コンストラクタに引数を指定できないので、オブジェクトのインスタンスを
 * 外部からの値で初期化する必要がある場合は、このようにする。
 */
com.yscjp.jsapp.JSMouseEvent.prototype.initJSMouseEvent = function (e) {

    /*
     * 必ず継承元の初期化メソッドを、呼び出す必要がある。
     */
    this.initJSInputEvent(e);

    if ("undefined" != typeof this.nativeEvent.screenX) {

        this.screenX = this.nativeEvent.screenX;

    } else {
    }

    if ("undefined" != typeof this.nativeEvent.screenY) {

        this.screenY = this.nativeEvent.screenY;

    } else {
    }

    if ("undefined" != typeof this.nativeEvent.clientX) {

        this.clientX = this.nativeEvent.clientX;

    } else {
    }

    if ("undefined" != typeof this.nativeEvent.clientY) {

        this.clientY = this.nativeEvent.clientY;

    } else {
    }

    /*
     * CSSOM View Moduleの仕様を取り入れる。
     */
    if ("undefined" != typeof this.nativeEvent.offsetX) {

        /*
         * XがあればYもあると想定
         */
        this.offsetX = this.nativeEvent.offsetX;
        this.offsetY = this.nativeEvent.offsetY;

    } else if ("undefined" != typeof this.nativeEvent.layerX) {

        this.offsetX = this.nativeEvent.layerX;
        this.offsetY = this.nativeEvent.layerY;

    } else {

        this.offsetX = -1;
        this.offsetY = -1;
    }

    if ("undefined" != typeof this.nativeEvent.button) {

        /*
         * DOM Events Specificationを正しくサポートしている場合は、
         * そのまま設定する。
         * そうでないと思われる場合は、DOMの仕様にあわせる。
         * 但し、何を持って正しくサポートしていると判断するかは、
         * 非常に難しい。
         * ここでは、body要素にaddEventListenerメソッドが存在する場合、
         * 正しい値を設定していると思う。
         * このあたりは、まだまだ相当な考慮が必要。
         */
        if (document.body.addEventListener) {

            this.button = this.nativeEvent.button;

        } else {

            if (1 == this.nativeEvent.button) {

                /*
                 * 左ボタン押下
                 */
                this.button = 0;

            } else if (4 == this.nativeEvent.button) {

                /*
                 * 中ボタン押下
                 */
                this.button = 1;

            } else if (2 == this.nativeEvent.button) {

                /*
                 * 右ボタン押下
                 */
                this.button = 2;

            } else {

                /*
                 * 上記以外は、いろいろなボタンの組合せなので、
                 * 無視する。
                 * 左ボタンを押したと思う。
                 */
                this.button = 0;
            }
        }
    }

    if ("undefined" != typeof this.nativeEvent.relatedTarget) {

        this.relatedTarget = this.nativeEvent.relatedTarget;

    } else {

        switch (this.type) {
            case 'mouseout' :

                this.relatedTarget = this.nativeEvent.toElement;
                break;

            case 'mouseover' :

                this.relatedTarget = this.nativeEvent.fromElement;
                break;

            default :

                this.relatedTarget = null;
                break;
        }
    }
};


/******************************************************************************
 * キーボードイベント
 * DOM Events Specificationの、KeyboardEventインターフェイスに該当する。
 *****************************************************************************/
com.yscjp.jsapp.JSKeyboardEvent = function(e) {

//    /* イベント発行の元となったキーのユニコード値 */
//    this.keyIdentifier = null;

    if (e) {

        this.initJSKeyboardEvent(e);
    }
};
com.yscjp.jsapp.JSKeyboardEvent.prototype = new com.yscjp.jsapp.JSInputEvent();

/**
 * インスタンスの初期化
 * コンストラクタに引数を指定できないので、オブジェクトのインスタンスを
 * 外部からの値で初期化する必要がある場合は、このようにする。
 */
com.yscjp.jsapp.JSKeyboardEvent.prototype.initJSKeyboardEvent = function (e) {

    /*
     * 必ず継承元の初期化メソッドを、呼び出す必要がある。
     */
    this.initJSInputEvent(e);

    if (this.nativeEvent.which) {

        this.keyIdentifier = e.which;
        this.keyCode = e.which;

    } else if (this.nativeEvent.keyCode) {

        this.keyIdentifier = e.keyCode;
        this.keyCode = e.keyCode;
    }
};

