/**
 * HTMLとJavaScriptによる、編集可能なテーブルコンポーネント
 * テーブルモデルとフィルタモデルのデフォルト実装
 *
 * Version 0.0.0.1
 *
 * このプログラムのライセンスについて。
 * このプログラムは、MITライセンス(The MIT License)の元で公開します。
 * Released under the MIT Licenses.
 *
 * ライセンスの原文は、以下のURLを参照してください。
 * http://www.yscjp.com/licenses/MIT_LICENSE.txt
 *
 * Copyright (c) 2010-2012 Yagyu Teru, http://yscjp.com/
 *
 */

var com;
if (!com) com = {};
if (!com.yscjp) com.yscjp = {};
if (!com.yscjp.jsapp) com.yscjp.jsapp = {};

/*****************************************************************************
 * テーブルモデルイベント
 ****************************************************************************/
/**
 * テーブルモデルから、テーブルビューへ、
 * モデルの内容が変更されたことを通知するイベント。
 * コンストラクタ。
 * @param model : テーブルモデル
 * @param startRow : 数値 : 変更された最初の行(この行も含む)。
 *        この値がマイナスの場合、モデル全体が変更されたとして、
 *        テーブルの内容を、全て書き換える。
 * @param endRow : 数値 : 変更された最期の行(この行も含む)。
 * @param startColumn : 数値 : 変更された最初の列(この列も含む)。
 * @param endColumn : 数値 : 変更された最期の列(この列も含む)。
 * @param type : 数値 : 変更された最期の列(この列も含む)。
 */
com.yscjp.jsapp.JSTableModelEvent = function (model,
                                              startRow,
                                              endRow,
                                              startColumn,
                                              endColumn,
                                              type) {

    this.model       = model;
    this.startRow    = startRow;
    this.endRow      = endRow;
    this.startColumn = startColumn;
    this.endColumn   = endColumn;
    this.type        = type;
};

com.yscjp.jsapp.JSTableModelEvent.CELL_UPDATE = 1;
com.yscjp.jsapp.JSTableModelEvent.ROW_INSERT  = 2;
com.yscjp.jsapp.JSTableModelEvent.ROW_DELETE  = 3;
com.yscjp.jsapp.JSTableModelEvent.FOOTER_CELL_UPDATE = 4;

/*****************************************************************************
 * テーブルモデル
 ****************************************************************************/
com.yscjp.jsapp.JSTableModel = function () {};

(function() {

    var JSTableModel = com.yscjp.jsapp.JSTableModel;
    var JSTableModelEvent = com.yscjp.jsapp.JSTableModelEvent;

    JSTableModel.prototype.initJSTableModel = function () {

        this.listenerList = new Array();
        this.rows = new Array();
        this.displayRows = new Array();
        this.isEditableRows = new Array();
        this.columnCount = -1;
    };

    /**
     * テーブルの名前を返します。
     * @return 文字列 : テーブルの名前。この実装では空文字を返します
     */
    JSTableModel.prototype.getTableName = function () {

        return "";
    }

    /**
     * テーブルのキャプションを返します。
     * @return 文字列 : テーブルのキャプション。この実装では空文字を返します
     */
    JSTableModel.prototype.getTableCaption = function () {

        return "";
    }

    /**
     * ヘッダー部の行数を返します。
     * この実装は、固定で0を返します。
     */
    JSTableModel.prototype.getHeadRowCount = function () {

        return 0;
    };

    /**
     * ヘッダー部の列数を返します。
     * このは、ボディ部と同じ値を返します。
     */
    JSTableModel.prototype.getHeadColumnCount = function () {

        return this.columnCount;
    };

    /**
     * フッター部の行数を返します。
     * この実装は、固定で0を返します。
     */
    JSTableModel.prototype.getFootRowCount = function () {

        return 0;
    };

    /**
     * フッター部の列数を返します。
     * このは、ボディ部と同じ値を返します。
     */
    JSTableModel.prototype.getFootColumnCount = function () {

        return this.columnCount;
    };

    /**
     * ボディ部の行数を返す。
     */
    JSTableModel.prototype.getRowCount = function () {

        return this.rows.length;
    };

    /**
     * 列数を返す。
     * このメソッドが-1を返すと、テーブルビューは、
     * 行毎に異なる列数で、テーブルを作成する。
     */
    JSTableModel.prototype.getColumnCount = function () {

        return this.columnCount;
    };

    /**
     * 指定行の列数を返す。
     * @param rowIndex : 数値 : 行インデックス。
     */
    JSTableModel.prototype.getColumnCountAt = function (rowIndex) {

        if (-1 != this.columnCount) {

            return this.columnCount;
        }

        var values = this.rows[rowIndex];

        if (values) {

            return values.length;
        }

        return 0;
    };

    /**
     * 列のタイトルを返す。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 文字列 : 指定カラムのタイトル。
     */
    JSTableModel.prototype.getColumnTitleAt = function (columnIndex) {

        return "" + columnIndex;
    };

    /**
     * 列の名前を返す。
     * この名前は、表示上は使用されないが、列のクラス名称として使用される。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 文字列 : 指定カラムの名称。
     */
    JSTableModel.prototype.getColumnNameAt = function (columnIndex) {

        return "";
    };

    /**
     * リスナの登録
     * モデルの内容が変更されたことを通知する、リスナを登録する。
     * リスナは、以下のメソッドを実装している必要がある。
     *  tableChanged(JSTableModelEvent)
     * @param view : Object : リスナ
     */
    JSTableModel.prototype.addTableModelListener = function (view) {

        if (! view) {

            throw "com.yscjp.jsapp.JSTableModel#addTableModelListener:"+
                "IllegalArgument:view is null.";
        }

        if (! view.tableChanged) {

            throw "com.yscjp.jsapp.JSTableModel#addTableModelListener:"+
                "NoSuchMethod:view.tableChanged";
        }

        this.listenerList.push(view);

        this.fireTableStructureChanged();
    };

    /**
     * モデルの内容が変更されたことを、全てのリスナーに通知する。
     * @param rowIndex : 数値 : 値が変更された行インデックス。
     * @param columnIndex : 数値 : 値が変更された列インデックス。
     */
    JSTableModel.prototype.fireTableChanged = function (modelEvent) {

        for (var i = 0; i < this.listenerList.length; i ++) {

            var view = this.listenerList[i];

            if (view) {

                var newEvent = null;

                if (null != modelEvent) {

                    /*
                     * viewがフィルターであった場合など、
                     * その中でイベントを書き換えたしまうので、
                     * それが他のリスナに影響しないように、
                     * 新たなイベントインスタンスを作成して渡す。
                     * 本来ここでやるべきではないかもしれないが、
                     * ここでやれば間違いない。
                     */
                    newEvent = new JSTableModelEvent(modelEvent.model,
                                                     modelEvent.startRow,
                                                     modelEvent.endRow,
                                                     modelEvent.startColumn,
                                                     modelEvent.endColumn,
                                                     modelEvent.type);
                }

                view.tableChanged(newEvent);
            }
        }
    };

    /**
     * 指定のセルの内容が変更されたことを、全てのリスナーに通知する。
     * @param rowIndex : 数値 : 値が変更された行インデックス。
     * @param columnIndex : 数値 : 値が変更された列インデックス。
     */
    JSTableModel.prototype.fireTableCellUpdated = function (rowIndex, columnIndex) {

        var modelEvent = new JSTableModelEvent(this,
                                               rowIndex,
                                               rowIndex,
                                               columnIndex,
                                               columnIndex,
                                               JSTableModelEvent.CELL_UPDATE);

        this.fireTableChanged(modelEvent);
    };

    /**
     * 指定の行が追加されたことを、全てのリスナーに通知する。
     * @param rowIndex : 数値 : 追加された行インデックス。
     */
    JSTableModel.prototype.fireTableRowInserted = function (rowIndex) {

        var modelEvent = new JSTableModelEvent(this,
                                               rowIndex,
                                               -1,
                                               -1,
                                               -1,
                                               JSTableModelEvent.ROW_INSERT);

        this.fireTableChanged(modelEvent);
    };

    /**
     * 指定の行が削除されたことを、全てのリスナーに通知する。
     * @param rowIndex : 数値 : 削除された行インデックス。
     */
    JSTableModel.prototype.fireTableRowDeleted = function (rowIndex) {

        var modelEvent = new JSTableModelEvent(this,
                                               rowIndex,
                                               -1,
                                               -1,
                                               -1,
                                               JSTableModelEvent.ROW_DELETE);

        this.fireTableChanged(modelEvent);
    };

    /**
     * モデルの構造が変更されたことを、全てのリスナーに通知する。
     */
    JSTableModel.prototype.fireTableStructureChanged = function () {

        this.fireTableChanged(null);
    };

    /**
     * 指定セルの値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値
     */
    JSTableModel.prototype.getValueAt = function (rowIndex, columnIndex) {

        var values = this.rows[rowIndex];

        if (values) {

            var value = values[columnIndex];
            if (value) {

                return value;
            }
        }

        return "";
        //return this.getEditableValueAt(rowIndex, columnIndex);
    };

    /**
     * 指定セルの表示用の値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値を表示用に編集した文字列。
     */
    JSTableModel.prototype.getDisplayValueAt = function (rowIndex, columnIndex) {

        var values = this.displayRows[rowIndex];

        if (values) {

            var value = values[columnIndex];
            if (value) {

                return value;
            }
        }

        return this.getValueAt(rowIndex, columnIndex);
    };

    /**
     * フッターの指定セルの値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値
     */
    JSTableModel.prototype.getFootValueAt = function (rowIndex, columnIndex) {

        return "";
    };

    /**
     * フッターの指定セルの表示用の値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値を表示用に編集した文字列。
     */
    JSTableModel.prototype.getFootDisplayValueAt = function (rowIndex, columnIndex) {

        return this.getFootValueAt(rowIndex, columnIndex);
    };

    /**
     * 指定の値を、モデルにセットする。
     * @param value : Object : セットする値。
     * @param rowIndex : 数値 : 値をセットする行数。
     * @param columnIndex : 数値 : 値をセットする列数。
     */
    JSTableModel.prototype.setValueAt = function (value, rowIndex, columnIndex) {

        var values = this.rows[rowIndex];
        values[columnIndex] = value;

        this.fireTableCellUpdated(rowIndex, columnIndex);
    };

    /**
     * 指定の値を、モデルにセットする。
     * @param value : Object : セットする値。
     * @param rowIndex : 数値 : 値をセットする行数。
     * @param columnIndex : 数値 : 値をセットする列数。
     */
    JSTableModel.prototype.setDisplayValueAt = function (value, rowIndex, columnIndex) {

        var values = this.displayRows[rowIndex];
        if (!values) {
            values = new Array();
            this.displayRows[rowIndex] = values;
        }
        values[columnIndex] = value;

        this.fireTableCellUpdated(rowIndex, columnIndex);
    };

    JSTableModel.prototype.setCellEditable = function (isCellEditable, rowIndex, columnIndex) {

        var values = this.isEditableRows[rowIndex];

        if (!values) {

            values = new Array();
            this.isEditableRows[rowIndex] = values;
        }

        values[columnIndex] = isCellEditable;

        this.fireTableCellUpdated(rowIndex, columnIndex);
    };

    /**
     * 編集可否判定
     * 指定セルが編集可能かどうかを返す。
     * @param rowIndex : 数値 : 行インデックス。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 必ずtrue
     */
    JSTableModel.prototype.isCellEditable = function (rowIndex, columnIndex) {

        var values = this.isEditableRows[rowIndex];

        if (values) {

            var value = values[columnIndex];
            if (!value) {

                return value;
            }
        }

        return true;
    };

    /**
     * 指定の行に、空の行を挿入する。
     * @param rowIndex : 数値 : 挿入する行インデックス。
     *                   -1を指定すると、最期に行を追加する。
     * @param values : 配列 : 新規行の内容の配列。
     */
    JSTableModel.prototype.insertRow = function (rowIndex) {

        var values = new Array();
        this.rows.splice(rowIndex, 0, values);
        this.fireTableRowInserted(rowIndex);
    };

    /**
     * 指定のインッデクスの行を削除する。
     * @param rowIndex : 数値 : 削除する行インデックス。
     */
    JSTableModel.prototype.deleteRow = function (rowIndex) {

        var values = new Array();
        this.rows.splice(rowIndex, 1);
        this.fireTableRowDeleted(rowIndex);
    };

    /**
     * 指定されたインデックスが有効かを判定を行います。
     * 指定されたインデクスが有効の場合true、有効ではない場合falseを返します。
     * @param rowIndex : 数値 : 行インデックス。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 指定されたインデクスが有効の場合true、有効ではない場合false
     */
    JSTableModel.prototype.isIndexValid = function (rowIndex, columnIndex) {
        //var Log = com.yscjp.jsapp.util.Log;
        //Log.debug("this.getRowCount():" + this.getRowCount());
        //Log.debug("this.getColumnCountAt():" + this.getColumnCountAt(rowIndex));
        /*
         * このモデルが保持する行数より大きな行インデックスは無効
         */
        if (this.getRowCount() <= rowIndex) {

            return false;
        }

        /*
         * 指定行の列数より大きな列インデックスは無効
         */
        if (this.getColumnCountAt(rowIndex) <= columnIndex) {

            return false;
        }

        return true;
    };
})();


/*****************************************************************************
 * デフォルトテーブルモデル
 ****************************************************************************/
//com.yscjp.jsapp.DefaultTableModel = {};

/*
 * コンストラクタ
 * @param rowCount 行数
 * @param columnCount 列数
 */
com.yscjp.jsapp.DefaultTableModel = function (rowCount, columnCount) {

    this.initJSTableModel();
    this.columnCount = columnCount;

    for (var i = 0; i < rowCount; i++) {

        if (-1 == this.columnCount) {

            this.rows[i] = new Array();

        } else {

            this.rows[i] = new Array(columnCount);
        }
    }
};

com.yscjp.jsapp.DefaultTableModel.prototype = new com.yscjp.jsapp.JSTableModel();

com.yscjp.jsapp.DefaultTableModel.createTableModel = function (arg0, arg1) {

    var tableModel;

    if (!arg0) {

        throw "com.yscjp.jsapp.DefaultTableModel#createTableModel:Argument is null.";

    } else if (!arg1) {

        /*
         * 引数が1つの場合は、テーブル要素が引数なので、
         * その行数とカラム数でテーブルモデルを初期化する。
         */
        var htmlTable = arg0;

        /*
         * テーブルのボディ部の行数を取得する。
         */
        var rowCount = htmlTable.getBodiesRowCount();

        /*
         * 列数は、ボディ部の先頭行の列数とする。
         * 取り合えずは、行が存在しないなどというイレギュラは考慮しない TODO
         */
        var headColumnCount = htmlTable.getHeadColumnCount();
        var columnCount = htmlTable.getRowFromBodiesAt(0).cells.length - headColumnCount;
        var tableModel = new com.yscjp.jsapp.DefaultTableModel(rowCount, columnCount);
        for (var i = 0; i < rowCount; i++) {

            var row = htmlTable.getRowFromBodiesAt(i);
            tableModel.rows[i] = new Array(row.cells.length);
            for (var j = 0; j < row.length; j++) {

                var cellText = DomUtil.getTextValueByNode(row.item(j));
                tableModel.setValueAt(cellText, i, j);
            }
        }
    } else {

        /*
         * 引数が2つの場合は行数と列数が指定された。
         */
        var rowCount = arg0;
        var columnCount = arg1;
        var tableModel = new com.yscjp.jsapp.DefaultTableModel(rowCount, columnCount);
    }

    return tableModel;
};

/*****************************************************************************
 * フィルターモデル
 * ビューとモデルの間で、列インデックスの変換などを行う。
 * ビューに対してはモデルとして、モデルに対してはビューとして
 * 見せかける。
 ****************************************************************************/
com.yscjp.jsapp.FilterModel = {};
com.yscjp.jsapp.FilterModel.prototype = new com.yscjp.jsapp.JSTableModel();

(function() {

    var FilterModel = com.yscjp.jsapp.FilterModel;

    /*
     * コンストラクタ
     * @param model : テーブルモデル : 実際にデータを格納するモデル
     */
    FilterModel = function () {

        //this.orgModel = null;
    };

    /*
     * テーブルモデルとして、実装が必須のメソッド
     */
    /**
     * 行数を返す。
     */
    FilterModel.prototype.getRowCount = function () {

        return this.orgModel.getRowCount();
    };

    /**
     * 列数を返す。
     * このメソッドが-1を返すと、テーブルビューは、
     * 行毎に異なる列数で、テーブルを作成する。
     */
    FilterModel.prototype.getColumnCount = function () {

        return this.orgModel.getColumnCount();
    };

    /**
     * 指定行の列数を返す。
     * @param rowIndex : 数値 : 行インデックス。
     */
    FilterModel.prototype.getColumnCountAt = function (rowIndex) {

        return this.orgModel.getColumnCountAt(rowIndex);
    };

    /**
     * 列のタイトルを返す。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 文字列 : 指定カラムのタイトル。
     */
    FilterModel.prototype.getColumnTitleAt = function (columnIndex) {

        return this.orgModel.getColumnTitleAt(columnIndex);
    };

    /**
     * リスナの登録
     * モデルの内容が変更されたことを通知する、リスナを登録する。
     * リスナは、以下のメソッドを実装している必要がある。
     *  tableChanged(JSTableModelEvent)
     * @param view : Object : リスナ
     */
    FilterModel.prototype.addTableModelListener = function (view) {

        this.view = view;
        this.orgModel.addTableModelListener(this);
    };

    /**
     * 指定セルの表示用の値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値
     */
    FilterModel.prototype.getValueAt = function (rowIndex, columnIndex) {

        var newIndices = this.changeIndex(rowIndex, columnIndex);
        rowIndex = newIndices[0];
        columnIndex = newIndices[1];
        return this.orgModel.getValueAt(rowIndex, columnIndex);
    };

    /**
     * 指定セルの編集用の値を取得する。
     * @param rowIndex : 数値 : 値を取得するセルの行インデックス。
     * @param columnIndex : 数値 : 値を取得するセルの列インデックス。
     * @return モデルが格納している値。
     */
    FilterModel.prototype.getEditableValueAt = function (rowIndex, columnIndex) {

        var newIndices = this.changeIndex(rowIndex, columnIndex);
        rowIndex = newIndices[0];
        columnIndex = newIndices[1];
        return this.orgModel.getEditableValueAt(rowIndex, columnIndex);
    };

    /**
     * 指定の値を、モデルにセットする。
     * @param value : Object : セットする値。
     * @param rowIndex : 数値 : 値をセットする行数。
     * @param columnIndex : 数値 : 値をセットする列数。
     */
    FilterModel.prototype.setValueAt = function (value, rowIndex, columnIndex) {

        var newIndices = this.changeIndex(rowIndex, columnIndex);
        rowIndex = newIndices[0];
        columnIndex = newIndices[1];
        this.orgModel.setValueAt(value, rowIndex, columnIndex);
    };

    /**
     * 編集可否判定
     * 指定セルが編集可能かどうかを返す。
     * @param rowIndex : 数値 : 行インデックス。
     * @param columnIndex : 数値 : 列インデックス。
     * @return this.orgModel.isCellEditableの戻り値
     */
    FilterModel.prototype.isCellEditable = function (rowIndex, columnIndex) {

        var newIndices = this.changeIndex(rowIndex, columnIndex);
        rowIndex = newIndices[0];
        columnIndex = newIndices[1];
        return this.orgModel.isCellEditable(rowIndex, columnIndex);
    };

    /**
     * インデックス変換
     * ビューのインデックスを、データモデルのインデックスに変換する。
     * @param rowIndex : 数値 : ビューの行インデックス。
     * @param columnIndex : 数値 : ビューの列インデックス。
     * @return : 配列 : １番目の要素を行インデックス、
     *                  ２番目の要素を列インデックスとする配列。
     */
    FilterModel.prototype.changeIndex = function (rowIndex, columnIndex) {

        return [rowIndex, columnIndex];
    };

    /*
     * ビューとして、実装が必須のメソッド
     */
    FilterModel.prototype.tableChanged = function (modelEvent) {

        if (null != modelEvent && 0 <= modelEvent.startRow) {

            modelEvent = this.changeEvent(modelEvent);
        }
        this.view.tableChanged(modelEvent);
    };

    /**
     * イベントの変換
     * イベントが指すインデックスを変換し返す。
     */
    FilterModel.prototype.changeEvent = function (modelEvent) {

        return modelEvent;
    };
})();

com.yscjp.jsapp.util.StackTrace.insert(com.yscjp.jsapp.JSTableModel, "com.yscjp.jsapp.JSTableModel");
//com.yscjp.jsapp.util.StackTrace.insert(com.yscjp.jsapp.DefaultTableModel, "com.yscjp.jsapp.DefaultTableModel");
com.yscjp.jsapp.util.StackTrace.insert(com.yscjp.jsapp.FilterModel, "com.yscjp.jsapp.FilterModel");
