/**
 * HTMLのtable要素を、javascriptで操作するクラスのベース機能を実装するクラスです。
 * tableへの入力フォーカスの管理とセルの選択及び、テーブルモデルとの表示内容の同期をサポートします。
 *
 * Version 00.12.00.00
 *
 * このプログラムのライセンスについて。
 * このプログラムは、MITライセンス(The MIT License)の元で公開します。
 * Released under the MIT Licenses.
 *
 * ライセンスの原文は、以下のURLを参照してください。
 * http://www.yscjp.com/licenses/MIT_LICENSE.txt
 *
 * Copyright (c) 2010-2013 Yagyu Teru, http://yscjp.com/
 *
 */
var com;
if (!com) com = {};
if (!com.yscjp) com.yscjp = {};
if (!com.yscjp.jsapp) com.yscjp.jsapp = {};

/**
 * テーブルのベースとなるクラスのコンストラクタ。
 * 継承時にこのコンストラクタが実行されるので、この中ではフィールドの初期化のみを行う。
 */
com.yscjp.jsapp.JSTableBase = function() {

    /* プライベート変数の定義 -------------------------------------*/
    /* 処理対象のテーブル要素 */
    this.tableElement;
    this.tableTopPosition;
    this.tableLeftPosition;
    /* テーブルオブジェクト */
    this.htmlTable;
    //this.cellOver;

    /** テーブルモデル */
    this.tableModel;

    /** プラグインを格納した配列 */
    this.plugins;

    /** 選択された範囲の左上のセル */
    this.selectedCellLeftTop;

    /** 選択された範囲の右下のセル */
    this.selectedCellRightBottom;

    /** ヘッダ列のサイズ。 */
    this.headColumnCount;

    /**
     * 現在選択されているセルの列インデックス。
     * この値は、テーブル要素の物理的なインデックスなので、ヘッダ列も含みます。
     */
    this.currentColumn = 0;

    /**
     * 現在フォーカスが存在するセルの列インデックス。
     * キーボードで直下に移動した場合、そのセルが移動不可（編集不可）の場合に、
     * その前後に一時的に移動する場合に、currentColumnと値が異なる。
     * 通常は、currentColumnと同じインデックスが格納されている。
     */
    this.focusedColumn = 0;

    /**
     * 現在選択されているセルの行インデックス。
     * この値は、テーブル要素のボディ部の物理的なインデックスなので、ヘッダ行は含みません。
     */
    this.currentRow = 0;

    /**
     * 現在フォーカスが存在するセル。
     * キーボードでフォーカス出来ないセルが選択されたような場合、
     * この内容はcurrentCellと異なるセルを指します。
     * 行単位にフォーカスできないような場合は、nullが設定される可能性もあります。
     */
    this.focusedCell;

    /**
     * 現在選択されているセル。
     * このセルは、フォーカスが可能かどうかに関わらず、迫っていされます。
     * キーボードによる選択セルの移動は、この内容を起点に行われます。
     * nullになることは、行列が存在する限りありえません。
     */
    this.currentCell;

    this.focusCtrl;
    this.selectedTopRow = 0;
    this.selectedBottomRow = 0;
    this.selectedLeftColumn = 0;
    this.selectedRightColumn = 0;

    /**
     * テーブルモデルで指定されたクラス名。
     * 対象テーブル要素のclass属性に設定したもの。
     */
    this.modelClassName = "";

    /*
     * テーブルモデルが変更され多様な場合に、物理的な構造を変更する必要があるかを示すフラグ。
     * 画面上のHTMLの構成をそのまま使用する場合、このフラグはfalseとなり、
     * initializeViewメソッドは、何もしないで処理を終了します。
     */
    this.isPhysicalStructureChange = true;
    this.rowChangeListeners = new Array();
};

(function() {

    /*
     * モジュールのインポート
     */
    var Log = com.yscjp.jsapp.util.Log;
    var JSEvent = com.yscjp.jsapp.JSEvent;
    var Table = com.yscjp.jsapp.html.Table;
    var HtmlUtil = com.yscjp.jsapp.util.JSHtmlUtil;
    var DomUtil = com.yscjp.jsapp.util.DomUtil;
    var JSTableBase = com.yscjp.jsapp.JSTableBase;
    var JSTableModelEvent = com.yscjp.jsapp.JSTableModelEvent;

    /**
     * このテーブルのヘッダ列数を返します。
     */
    JSTableBase.prototype.getHeadColumnCount = function () {

        return this.headColumnCount;
    };

    /**
     * ヘッダ列数を設定します。
     * ヘッダ列数の変更は構造の変更を意味するので、現在列のインデックスにヘッダ列数を設定します。
     */
    JSTableBase.prototype.setHeadColumnCount = function (headColumnCount) {

        this.headColumnCount = headColumnCount;
        this.setSelectedColumnIndex(this.headColumnCount);
    };

    /**
     * テーブルの初期化を行います。
     * JSTableを使用する場合は、JSTableを継承したクラスを作成し、
     * その中でこのメソッドを呼び出す必要があります。
     * このコンストラクタに指定できる引数の内容は、以下の組み合わせである必要があります。
     * @param id
     * 指定のidに該当するtable要素を処理します。行数列数やセルの内容は、HTMLに記述された内容に従います。
     * @param id
     * @param initFrom
     * 指定のidに該当するtable要素を処理します。行数列数やセルの内容は、HTMLに記述された内容に従います。
     * @param id
     * @param initFrom
     * @param rowNum
     * @param colNum
     * 指定のidに該当するtable要素を処理します。行数列数は指定のサイズとなります。セルの内容は、全て空となります。
     * @param id
     * @param initFrom
     * @param this.tableModel
     * 指定のidに該当するtable要素を処理します。行数列数やセルの内容は、this.tableModelの内容に従います。
     */
    JSTableBase.prototype.initJSTableBase = function (id, initFrom, arg0, arg1) {

        /* 処理対象のテーブル要素を取得します。 */
        this.tableElement = document.getElementById(id);
        this.tableTopPosition = this.tableElement.offsetTop;
        this.tableLeftPosition = this.tableElement.offsetLeft;

        /* テーブルのキーボードイベントを処理するテキストエリアを追加する。 */
        var ta = document.createElement('textarea');
        this.tableElement.style.position = 'relative';
        ta = document.body.appendChild(ta);

        if (com.yscjp.jsapp.util.UA.isFixedSupport()) {

            ta.style.position = 'fixed';

        } else {

            ta.style.position = 'absolute';
        }

        ta.style.top = '-100px';
        ta.style.left = '-100px';
        ta.style.width = '20px';
        ta.style.height = '20px';
        ta.style.border = '0px';
        ta.id = id + '_ctrl';
        this.focusCtrl = ta;

        /*
         * テーブルオブジェクトを作成する。
         * この時、idで指定された要素が不正の場合、例外が投げられる。
         */
        this.htmlTable = new Table(this.tableElement);
        this.setHeadColumnCount(this.htmlTable.getHeadColumnCount());
        this.currentCell = this.htmlTable.getCellFromBodiesAt(this.getSelectedRowIndex(), this.getSelectedColumnIndex());

        if (null == initFrom) {

            /*
             * id だけ指定された。
             */
            this.isPhysicalStructureChange = false;
            var model = com.yscjp.jsapp.DefaultTableModel.createTableModel(this.htmlTable);
            this.setTableModel(model);

        } else if (null == arg0) {

            /*
             * id と initFrom が指定された。
             */
        } else if (null == arg1) {

            /*
             * id と initFrom と arg0 が指定された。
             */
            this.setTableModel(arg0);
            //this.initializeView();

        } else {
        }

        var _this = this;

        function tableResize(e) {

            var topPosition = _this.tableTopPosition + _this.currentCell.offsetTop;
            var leftPosition = _this.tableLeftPosition + _this.currentCell.offsetLeft;
            var width = _this.currentCell.offsetWidth;
            var height = _this.currentCell.offsetHeight;
        }

        JSEvent.addEventListener(window, "resize", tableResize, false);
    };

    /**
     * テーブル表示の初期化を行う。
     * この実装では、tableModel の内容に合わせてViewを初期化します。
     * Modelからも呼び出される。
     */
    JSTableBase.prototype.initializeView = function () {

        if (null == this.tableModel) {

            return;
        }

        if (!this.isPhysicalStructureChange) {

            /*
             * 物理的な変更を行わないテーブルの場合は、
             * 何もしないで処理を終了する。
             */
            return;
        }

        /*
         * 一旦全ての行を削除する。
         */
        this.removeAllRow();

        this.focusedColumn = 0;
        this.currentRow = 0;
        this.currentColumn = 0;
        this.selectedTopRow = 0;
        this.selectedBottomRow = 0;
        this.selectedLeftColumn = 0;
        this.selectedRightColumn = 0;

        /*
         * テーブルに関する設定
         */
        if (0 != this.modelClassName.length) {

            HtmlUtil.removeClassValue(this.tableElement, this.modelClassName);
        }
        this.modelClassName = this.tableModel.getTableName();
        if (0 != this.modelClassName.length) {

            HtmlUtil.addClassValue(this.tableElement, this.modelClassName);
        }
        this.htmlTable.setCaption(this.tableModel.getTableCaption());
        this.initHeadColumn();

        /*
         * ヘッダ部の初期化。
         */
        var headRowCount = this.tableModel.getHeadRowCount();
        /* 以下のgetHeadColumnCount()が返すのは、ヘッダ行の列数なので注意すること。 */
        var headColumnCount = this.tableModel.getHeadColumnCount();
        for (var i = 0; i < headRowCount; i ++) {

            var row = this.htmlTable.insertHeadRowAt(-1);
            this.addHeadColumn(row);
            for (var j = 0; j < headColumnCount; j ++) {

                var cell = row.insertCell(-1);
                if (0 == i) {

                    /*
                     * テーブルモデルは、複数行のヘッダをサポートしていないので、
                     * 先頭行にのみ、タイトルを設定する。
                     */
                    DomUtil.setTextValue(cell, this.tableModel.getColumnTitleAt(j));
                    HtmlUtil.addClassValue(cell, this.tableModel.getColumnNameAt(j));

                }
            }
        }

        var rowCount = this.tableModel.getRowCount();
        var columnCount = this.tableModel.getColumnCount();
        for (var i = 0; i < rowCount; i ++) {

            var columnCountInRow = columnCount;
            if (-1 == columnCount) {

                /*
                 * 行毎に列数が異なる。
                 */
                columnCountInRow = this.tableModel.getColumnCountAt(i);
            }

            var row = this.htmlTable.insertRowAt(-1);
            this.addHeadColumn(row);
            for (var j = 0; j < columnCountInRow; j ++) {

                var cell = row.insertCell(-1);
                HtmlUtil.addClassValue(cell, this.tableModel.getColumnNameAt(j));
            }
        }

        /*
         * フッタ部の初期化。
         */
        var footRowCount = this.tableModel.getFootRowCount();
        var footColumnCount = this.tableModel.getFootColumnCount();
        for (var i = 0; i < footRowCount; i ++) {

            var row = this.htmlTable.insertFootRowAt(-1);
            this.addHeadColumn(row);
            for (var j = 0; j < footColumnCount; j ++) {

                var cell = row.insertCell(-1);
            }
        }
    };

    /**
     * デフォルトでは、各行の先頭にヘッダ列が付加されます。
     * これを回避するためには、このメソッドを再定義して、何もしない実装にしてください。
     */
    JSTableBase.prototype.initHeadColumn = function () {

        this.setHeadColumnCount(1);
        var col = this.htmlTable.getColAt(0);
        HtmlUtil.addClassValue(col, "com_yscjp_htmlclass_headcolumn");
    };

    /**
     * デフォルトでは、各行の先頭にヘッダ列が付加されます。
     * これを回避するためには、このメソッドを再定義して、何もしない実装にしてください。
     */
    JSTableBase.prototype.addHeadColumn = function (row) {
        row.insertCell(-1);
    };

    /**
     * このオブジェクトのViewとなる表から、全ての行を削除します。
     */
    JSTableBase.prototype.removeAllRow = function () {

        this.htmlTable.removeAllRow();
    };

    /**
     * このテーブルにプラグインを追加します。
     */
    JSTableBase.prototype.addPlugin = function (plugin) {

        plugin.init(this);
    };

    /**
     * このテーブルに、テーブルモデルを設定します。
     * @param this.tableModel テーブルモデル
     */
    JSTableBase.prototype.setTableModel = function (t) {

        if (null == t) {

            throw "com.yscjp.jsapp.JSTableBase#setTableModel:Argument is null.";
        }

        /*
         * 以下のメソッドを実装している必要がある。
         */
        if (!t.addTableModelListener) {throw "No such method.:addTableModelListener";}
        if (!t.getTableName) {throw "No such method.:getTableName";}
        if (!t.getTableCaption) {throw "No such method.:getTableCaption";}
        if (!t.getColumnCount) {throw "No such method.:getColumnCount";}
        if (!t.getColumnCountAt) {throw "No such method.:getColumnCountAt";}
        if (!t.getColumnTitleAt) {throw "No such method.:getColumnTitleAt";}
        if (!t.getColumnNameAt) {throw "No such method.:getColumnNameAt";}
        if (!t.getFootColumnCount) {throw "No such method.:getFootColumnCount";}
        if (!t.getFootRowCount) {throw "No such method.:getFootRowCount";}
        if (!t.getHeadColumnCount) {throw "No such method.:getHeadColumnCount";}
        if (!t.getHeadRowCount) {throw "No such method.:getHeadRowCount";}
        if (!t.getRowCount) {throw "No such method.:getRowCount";}
        if (!t.getValueAt) {throw "No such method.:getValueAt";}
        if (!t.isCellEditable) {throw "No such method.:isCellEditable";}
        if (!t.isIndexValid) {throw "No such method.:isIndexValid";}
        if (!t.setValueAt) {throw "No such method.:setValueAt";}

        this.tableModel = t;
        this.tableModel.addTableModelListener(this);

        /*
         * 初期表示時にフォーカスが当たるセルを設定する。
         */
        this.findCanFocusdCell(this.getSelectedRowIndex(), this.getSelectedColumnIndex());
    };

    /**
     * このテーブルのテーブルモデルを返します。
     * テーブルモデルが設定されていない場合、nullを返します。
     * @return this.tableModel テーブルモデル
     */
    JSTableBase.prototype.getTableModel = function () {

        return this.tableModel;
    };

    /**
     * テーブルモデルの内容が変更されたことを、テーブルに通知します。
     * TableModelから呼び出される。
     */
    JSTableBase.prototype.tableChanged = function (tableModelEvent) {

        if (null == this.tableModel) {

            return;
        }

        var startRow;
        var endRow;
        var startColumn;
        var endColumn;

        /*
         * イベントが指定されていないか、スタート行がマイナス指定の場合は、
         * テーブルを全面的に書き換える。
         */
        if (null == tableModelEvent || 0 > tableModelEvent.startRow) {

            this.initializeView();
            startRow = 0;
            endRow = this.htmlTable.getBodiesRowCount() - 1;
            startColumn = this.headColumnCount;
            endColumn = 99999;

        } else {

            /*
             * テーブルモデルのカラムインデックスは、ヘッダ列を考慮していないことに注意すること。
             */
            startRow    = tableModelEvent.startRow;
            endRow      = tableModelEvent.endRow;
            startColumn = tableModelEvent.startColumn + this.headColumnCount;
            endColumn   = tableModelEvent.endColumn + this.headColumnCount;
        }

        /*
         * 行の追加削除を行う。DELETE
         */
//        if (JSTableModelEvent.ROW_INSERT == tableModelEvent.type ||
//            JSTableModelEvent.ROW_DELETE == tableModelEvent.type) {
//        }

        for (var i = startRow; i <= endRow; i ++) {

            var row = this.htmlTable.getRowFromBodiesAt(i);
            var cellCount = (endColumn < row.cells.length) ? endColumn : (row.cells.length - 1);
            for (var j = startColumn; j <= cellCount; j ++) {

                var columnIndex = j;

                setViewValue(this, i, columnIndex);
                var cell = this.htmlTable.getCellFromBodiesAt(i, columnIndex);

                if (this.isCellEditable(i, columnIndex)) {

                    //this.setEditableValue(i, j);
                    HtmlUtil.removeClassValue(cell, "com_yscjp_htmlclass_nonEditableCell");
                    HtmlUtil.addClassValue(cell, "com_yscjp_htmlclass_editableCell");

                } else {

                    HtmlUtil.removeClassValue(cell, "com_yscjp_htmlclass_editableCell");
                    HtmlUtil.addClassValue(cell, "com_yscjp_htmlclass_nonEditableCell");
                }
            }
        }
    };

    /**
     * このテーブルオブジェクトが管理するテーブル要素を返します。
     */
    JSTableBase.prototype.getTableElement = function () {

        return this.tableElement;
    };

    /**
     * セルの選択が変更されたことを通知します。
     */
    JSTableBase.prototype.fireSelectChanged = function (rowIndex, columnIndex) {

        cellReleased(this);

        var currentRow = this.getSelectedRowIndex();

        this.setSelectedRowIndex(rowIndex);
        this.setSelectedColumnIndex(columnIndex);
        moveCurrentCell(this, this.getSelectedRowIndex(), this.getSelectedColumnIndex());

        cellSelected(this);

        runPluginMethod(this, "fireSelectChanged");
        //if (currentRow != this.getSelectedRowIndex()) {

            this.callRowChangeListeners(this.getSelectedRowIndex());
        //}
    };

    /**
     * 選択セルを指定のインデックスが指すセルに変更します。
     * このメソッドに渡されるインデックスは、既に矛盾がないこを確認されている必要があります。
     */
    function moveCurrentCell (_this, rowIndex, columnIndex) {

        _this.currentCell = _this.htmlTable.getCellFromBodiesAt(rowIndex, columnIndex);
        _this.focusedCell = null;
    }

    /**
     * セルの選択が開始されたことを通知します。
     */
    JSTableBase.prototype.fireSelectStart = function (rowIndex, columnIndex) {

        cellsRelease(this.htmlTable,
                     this.selectedTopRow,
                     this.selectedBottomRow,
                     this.selectedLeftColumn,
                     this.selectedRightColumn);

        this.startRow = rowIndex;
        this.startColumn = columnIndex;
        this.selectedTopRow = this.startRow;
        this.selectedBottomRow = this.startRow;
        this.selectedLeftColumn = this.startColumn;
        this.selectedRightColumn = this.startColumn;

        cellsSelect(this.htmlTable,
                    this.selectedTopRow,
                    this.selectedBottomRow,
                    this.selectedLeftColumn,
                    this.selectedRightColumn);
    };

    /**
     * セルの選択が開始されたことを通知します。
     */
    JSTableBase.prototype.fireSelectEnd = function (rowIndex, columnIndex) {

        cellsRelease(this.htmlTable,
                     this.selectedTopRow,
                     this.selectedBottomRow,
                     this.selectedLeftColumn,
                     this.selectedRightColumn);

        this.selectedTopRow = Math.min(this.startRow, rowIndex);
        this.selectedBottomRow = Math.max(this.startRow, rowIndex);
        this.selectedLeftColumn = Math.min(this.startColumn, columnIndex);
        this.selectedRightColumn = Math.max(this.startColumn, columnIndex);

        cellsSelect(this.htmlTable,
                    this.selectedTopRow,
                    this.selectedBottomRow,
                    this.selectedLeftColumn,
                    this.selectedRightColumn);
    };

    JSTableBase.prototype.getSelectedRowIndex = function () {

        return this.currentRow;
    };

    JSTableBase.prototype.setSelectedRowIndex = function (currentRow) {

        this.currentRow = currentRow;
    };

    JSTableBase.prototype.getSelectedColumnIndex = function () {

        return this.currentColumn;
    };

    JSTableBase.prototype.getFocusedColumnIndex = function () {

        return this.focusedColumn;
    };

    /**
     * 現在列インデックスに、指定の値を設定します。
     */
    JSTableBase.prototype.setSelectedColumnIndex = function (currentColumn) {

        this.currentColumn = currentColumn;
        this.focusedColumn = this.currentColumn;
    };

    /**
     * 現在フォーカスが当たっているセルの値を取得します。
     */
    JSTableBase.prototype.getValueCurrentCell = function () {

        return this.getValueAt(this.getSelectedRowIndex(), this.getSelectedColumnIndex());
    };

    /**
     * 現在フォーカスが当たっているセルに値を設定します。
     */
    JSTableBase.prototype.setValueCurrentCell = function (value) {

        this.setValueAt(value, this.getSelectedRowIndex(), this.getSelectedColumnIndex());
    };

    /**
     * 指定のセルの値を返します。
     * @param rowIndex 取得するセルの、テーブルボディ部の行インデックス。
     * @param columnIndex 取得するセルの、ヘッダ列を含んだ物理列インデックス。
     */
    JSTableBase.prototype.getValueAt = function (rowIndex, columnIndex) {

        return this.tableModel.getValueAt(rowIndex, columnIndex - this.headColumnCount);
    };

    /**
     * 指定のセルに値を設定します。
     * 編集不可の場合何もしません。
     * @param value 設定する値
     * @param rowIndex 設定するセルの、テーブルボディ部の行インデックス。
     * @param columnIndex 設定するセルの、ヘッダ列を含んだ物理列インデックス。
     */
    JSTableBase.prototype.setValueAt = function (value, rowIndex, columnIndex) {

        if (!this.isCellEditable(rowIndex, columnIndex)) {

            /*
             * 編集不可の場合は、何もしない。
             */
            return;
        }

        this.tableModel.setValueAt(value, rowIndex, columnIndex - this.headColumnCount);
    };

    /**
     * 指定のインデックスが、テーブルモデルで有効かを返します。
     * @param rowIndex 取得するセルの、テーブルボディ部の行インデックス。
     * @param columnIndex 取得するセルの、ヘッダ列を含んだ物理列インデックス。
     */
    JSTableBase.prototype.isIndexValid = function (rowIndex, columnIndex) {

        return this.tableModel.isIndexValid(rowIndex, columnIndex - this.headColumnCount);
    };

    /**
     * このテーブルにフォーカスを設定します。
     */
    JSTableBase.prototype.focus = function () {

        this.focusCtrl.focus();
    };

    JSTableBase.prototype.addRowChangeListener = function (listener) {
        this.rowChangeListeners.push(listener);
    };

    JSTableBase.prototype.callRowChangeListeners = function (rowIndex) {

        for (var i = 0; i < this.rowChangeListeners.length; i++) {

            this.rowChangeListeners[i](this, rowIndex);
        }
    };

    /**
     * カレント行を一行上げる。
     * カレント行が既に先頭行の場合何もしません。
     * @return 実際に移動した場合true、移動しなかった場合false
     */
    JSTableBase.prototype.goUp = function (evnt) {

        var rowIndex = this.getSelectedRowIndex();
        var columnIndex = this.getSelectedColumnIndex();

        rowIndex = this.rowChange(rowIndex, columnIndex, -1);
        if (-1 == rowIndex) {

            return false;
        }
        this.fireSelectChanged(rowIndex, columnIndex);  //  + this.headColumnCount

        return true;
    };

    /**
     * 選択行を一行下げる。
     * @return 実際に移動した場合true、移動しなかった場合false
     */
    JSTableBase.prototype.goDown = function (evnt) {

        var rowIndex = this.getSelectedRowIndex();
        var columnIndex = this.getSelectedColumnIndex();

        rowIndex = this.rowChange(rowIndex, columnIndex, 1);
        if (-1 == rowIndex) {

            return false;
        }

        this.fireSelectChanged(rowIndex, columnIndex);  //  + this.headColumnCount

        return true;
    };

    /**
     * 行の変更を行います。
     * 指定されたrowIndexとcolumnIndexを開始点として、フォーカスの設定できるセルを検索し、
     * 移動した行のインデックスを返します。
     * フォーカスの設定できるセルが存在しない場合、移動はせず-1を返します。
     * @param _this
     * @param rowIndex
     * @param columnIndex
     * @param direction 移動量。下へ移動する場合は1、上へ移動する場合は-1を指定します
     */
    JSTableBase.prototype.rowChange = function (rowIndex, columnIndex, direction) {

        var count = 1;
        while (1) {

            var index = count * direction;
            if (!this.htmlTable.isValidIndex(rowIndex + index, columnIndex)) {

                /*
                 * 移動した結果が有効なインデックスではない場合、-1を返す。
                 */
                return -1;
            }

            if (this.isCellFocus(rowIndex + index, columnIndex)) {

                return rowIndex + index;
            }

            count ++;
        }
    };

    /*
     * 左矢印キーが押下された。
     * @return 実際に移動した場合true、移動しなかった場合false
     */
    JSTableBase.prototype.goLeft = function (evnt) {

        var rowIndex = this.getSelectedRowIndex();
        var columnIndex = this.getSelectedColumnIndex();

        columnIndex = columnChange(this, rowIndex, columnIndex, -1);
        if (-1 == columnIndex) {

            return false;
        }

        this.fireSelectChanged(rowIndex, columnIndex);  //  + this.headColumnCount

        return true;
    };

    /*
     * 右矢印キーが押下された。
     * @return 実際に移動した場合true、移動しなかった場合false
     */
    JSTableBase.prototype.goRight = function (evnt) {

        var rowIndex = this.getSelectedRowIndex();
        var columnIndex = this.getSelectedColumnIndex();

        columnIndex = columnChange(this, rowIndex, columnIndex, 1);
        if (-1 == columnIndex) {

            return false;
        }

        this.fireSelectChanged(rowIndex, columnIndex);  //  + this.headColumnCount

        return true;
    };

    function columnChange(_this, rowIndex, columnIndex, direction) {

        var count = 1;
        while (1) {

            var index = count * direction;
            if (!_this.htmlTable.isValidIndex(rowIndex, columnIndex + index)) {

                /*
                 * 移動した結果が有効なインデックスではない場合、-1を返す。
                 */
                return -1;
            }

            if (_this.isCellFocus(rowIndex, columnIndex + index)) {

                return columnIndex + index;
            }

            count ++;
        }
    }

    /**
     * 指定のセルを開始点として、選択可能なセルを検索します。
     * 指定のセルが選択可能ではない場合、その行の後ろのセルを検索します。
     * 後ろに存在しない場合は、前を検索します。
     * その行に存在しない場合は、次の行を検索します。
     */
    JSTableBase.prototype.findCanFocusdCell = function (rowIndex, columnIndex) {

        var rowCount = this.htmlTable.getBodiesRowCount();
        for (var i = rowIndex; i < rowCount; i ++) {

            if (-1 != this.getCanFocusedColumn(i, columnIndex, true)) {
                return;
            }
        }
    };

    /**
     * 指定の行に選択可能なセルが存在するか検索します。
     */
    JSTableBase.prototype.getCanFocusedColumn = function (rowIndex, columnIndex, isfire) {

        isfire = isfire == null ? false : isfire;

        var row = this.htmlTable.getRowFromBodiesAt(rowIndex);
        var columnCount = row.cells.length;
        /* 後ろに検索。 */
        for (var j = columnIndex; j < columnCount; j ++) {
            if (this.isCellFocus(rowIndex, j)) {

                if (isfire) {

                    this.fireSelectChanged(rowIndex, j);
                }

                return j;
            }
        }
        /* 前に検索。 */
        for (var j = columnIndex - 1; j > 0; j --) {
            if (this.isCellFocus(rowIndex, j)) {

                if (isfire) {

                    this.fireSelectChanged(rowIndex, j);
                }

                return j;
            }
        }

        return -1;
    };

    /**
     * 指定のセルにフォーカスが設定できるかを返します。
     * @param rowIndex 取得するセルの、テーブルボディ部の行インデックス。
     * @param columnIndex 取得するセルの、ヘッダ列を含んだ物理列インデックス。
     */
    JSTableBase.prototype.isCellFocus = function (rowIndex, columnIndex) {

        if (!this.htmlTable.isValidIndex(rowIndex, columnIndex)) {

            /*
             * 移動した結果が有効なインデックスではない場合は、移動不可。
             */
            return false;
        }

        if (this.headColumnCount > columnIndex) {

            /*
             * ヘッダ列への移動は不可。
             */
            return false;
        }

        if (!this.isCellEditable(rowIndex, columnIndex)) {

            /*
             * 編集できない項目の場合は移動不可。
             */
            return false;
        }

        return true;
    };

    /**
     * 指定のセルが編集可能かを返します。
     * このメソッドに渡すインデックスは、ヘッダ列の数を考慮した、モデルのインデックスです。
     */
    JSTableBase.prototype.isCellEditable = function (rowIndex, columnIndex) {

        return this.tableModel.isCellEditable(rowIndex, columnIndex - this.headColumnCount);
    };

    /**
     * セルの表示内容を設定する。
     * このメソッドは、外部に公開しないprivateメソッドです。
     * @param rowIndex : 数値 : 行インデックス
     * @param columnIndex : 数値 : 列インデックス。表示上のヘッダ列を意識しないインデックスです
     */
    function setViewValue(_this, rowIndex, columnIndex) {

        var displayValue = _this.tableModel.getDisplayValueAt(rowIndex, columnIndex - _this.headColumnCount);
        var valueContainer = getCellValueContainerAt(_this, rowIndex, columnIndex);
        if (null == valueContainer) {
            return;
        }
//        if (displayValue.replace) {
//            displayValue = displayValue.replace(/\n/g, "<br />");
//            displayValue = "<doc>" + displayValue + "</doc>";
//            var newDoc = XmlUtil.perseFromString(displayValue);
//            DomUtil.removeAllChild(valueContainer);
//            DomUtil.deepClone(valueContainer, newDoc.documentElement);
//        }
//        else {
            DomUtil.setTextValue(valueContainer, displayValue);
//        }
    }

    /**
     * 指定セルの表示要素を返す。
     * アプリケーション固有の、セルビュー取得メソッドが定義されている場合は、
     * そちらを有効とする。
     * このメソッドは、外部に公開しないprivateメソッドです。
     * @param rowIndex : 数値 : 行インデックス。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 要素 : 指定のセルの表示要素。
     *         該当するセルが存在しない場合、null。
     *         デフォルトの表示要素が存在しない場合、セルを返す。
     */
    function getCellValueContainerAt(_this, rowIndex, columnIndex) {

        var cellValueContainer = null;

        /*
         * アプリケーション固有のセルビュー取得メソッドが定義されていた場合は、
         * 最初にそちらを呼び出す。
         */
//        if (getAppCellValueContainerAt) {
//
//            cellValueContainer = getAppCellValueContainerAt(rowIndex, columnIndex);
//
//            if (cellValueContainer) {
//
//                return cellValueContainer;
//            }
//        }

        var cell = getCellAt(_this, rowIndex, columnIndex);
        if (! cell) {

            return null;
        }

//        cellValueContainer =
//            DomUtil.getElementByAttributeValue(cell,
//                                               "className",
//                                               TableView.cellValueContainerClass);

        if (! cellValueContainer) {

            return cell;
        }

        return cellValueContainer;
    }

    /**
     * 指定のセルを返す。
     * このメソッドは、外部に公開しないprivateメソッドです。
     * @param rowIndex : 数値 : 行インデックス。
     * @param columnIndex : 数値 : 列インデックス。
     * @return 要素 : 指定のセル。該当するセルが存在しない場合、null。
     */
    function getCellAt(_this, rowIndex, columnIndex) {

        return _this.htmlTable.getCellFromBodiesAt(rowIndex, columnIndex);
    }

    /**
     * 登録されているすべてのプラグインの、指定のメソッドを実行します。
     * このメソッドは、外部に公開しないprivateメソッドです。
     */
    function runPluginMethod(_this, methodName) {

        for (var plugin in _this.plugins) {

            plugin[methodName]();
        }
    }

    /**
     * セルが選択されたときの処理を行う。
     * 選択されたセルに、選択された旨のクラスを設定する。
     * 選択された行のヘッダセルに、選択された旨のクラスを設定する。
     * 選択された列のヘッダセルに、選択された旨のクラスを設定する。
     * このメソッドは、外部に公開しないprivateメソッドです。
     */
    function cellSelected(_this) {

        if (_this.focusedCell) {

            HtmlUtil.addClassValue(_this.focusedCell, "com_yscjp_htmlclass_focusedCell");
            if (0 != _this.htmlTable.getHeadRowCount()) {

                /* ヘッダ行指定の場合にみ */
                var columnIndex = _this.getFocusedColumnIndex();
                HtmlUtil.addClassValue(_this.htmlTable.getCellFromHeadAt(0, columnIndex),
                                          "com_yscjp_htmlclass_focusedColumn");
            }
        } else {

            HtmlUtil.addClassValue(_this.htmlTable.getRowFromBodiesAt(_this.getSelectedRowIndex()),
                                      "com_yscjp_htmlclass_focusedRow");
        }

        HtmlUtil.addClassValue(_this.currentCell, "com_yscjp_htmlclass_currentCell");
        for (var i = 0; i < _this.htmlTable.getHeadRowCount(); i++) {

            /* ヘッダ行指定の場合にみ */
            var columnIndex = _this.getSelectedColumnIndex();
            HtmlUtil.addClassValue(_this.htmlTable.getCellFromHeadAt(i, columnIndex),
                                   "com_yscjp_htmlclass_currentColumn");
        }

        for (var i = 0; i < _this.headColumnCount; i ++) {

            /* ヘッダ列指定の場合にみ */
            HtmlUtil.addClassValue(_this.htmlTable.getCellFromBodiesAt(_this.getSelectedRowIndex(), i),
                                   "com_yscjp_htmlclass_currentRow");
        }
    }

    /**
     * セルの選択が解除されたときの処理を行う。
     * 選択されたセルから、選択された旨のクラスを削除する。
     * 選択された行のヘッダセルから、選択された旨のクラスを削除する。
     * 選択された列のヘッダセルから、選択された旨のクラスを削除する。
     * このメソッドは、外部に公開しないprivateメソッドです。
     */
    function cellReleased(_this) {

        if (_this.focusedCell) {

            HtmlUtil.removeClassValue(_this.focusedCell, "com_yscjp_htmlclass_focusedCell");
            if (0 != _this.htmlTable.getHeadRowCount()) {

                /* ヘッダ行指定の場合にみ */
                var columnIndex = _this.getFocusedColumnIndex();
                HtmlUtil.removeClassValue(_this.htmlTable.getCellFromHeadAt(0, columnIndex),
                                          "com_yscjp_htmlclass_focusedColumn");
            }

        }

        /* ヘッダ列指定の場合にみ */
        HtmlUtil.removeClassValue(_this.htmlTable.getRowFromBodiesAt(_this.getSelectedRowIndex()),
                                  "com_yscjp_htmlclass_focusedRow");

        HtmlUtil.removeClassValue(_this.currentCell, "com_yscjp_htmlclass_currentCell");
        for (var i = 0; i < _this.htmlTable.getHeadRowCount(); i++) {

            /* ヘッダ行指定の場合にみ */
            var columnIndex = _this.getSelectedColumnIndex();
            HtmlUtil.removeClassValue(_this.htmlTable.getCellFromHeadAt(i, columnIndex),
                                      "com_yscjp_htmlclass_currentColumn");
        }

        for (var i = 0; i < _this.headColumnCount; i ++) {

            /* ヘッダ列指定の場合にみ */
            HtmlUtil.removeClassValue(_this.htmlTable.getCellFromBodiesAt(_this.getSelectedRowIndex(), i),
                                      "com_yscjp_htmlclass_currentRow");
        }
    }

    /**
     * 指定の範囲のセルを選択状態にします。
     * このメソッドは、外部に公開しないprivateメソッドです。
     */
    function cellsSelect(htmlTable, selectedTopRow, selectedBottomRow, selectedLeftColumn, selectedRightColumn) {

        for (var rowIndex = selectedTopRow; rowIndex <= selectedBottomRow; rowIndex ++) {
            for (var columnIndex = selectedLeftColumn; columnIndex <= selectedRightColumn; columnIndex ++) {
                HtmlUtil.addClassValue(htmlTable.getCellFromBodiesAt(rowIndex, columnIndex),
                                       "com_yscjp_htmlclass_selectedCell");
            }
        }
    }

    /**
     * 指定の範囲のセルを選択解除状態にします。
     * このメソッドは、外部に公開しないprivateメソッドです。
     */
    function cellsRelease(htmlTable, selectedTopRow, selectedBottomRow, selectedLeftColumn, selectedRightColumn) {

        for (var rowIndex = selectedTopRow; rowIndex <= selectedBottomRow; rowIndex ++) {
            for (var columnIndex = selectedLeftColumn; columnIndex <= selectedRightColumn; columnIndex ++) {
                HtmlUtil.removeClassValue(htmlTable.getCellFromBodiesAt(rowIndex, columnIndex),
                                          "com_yscjp_htmlclass_selectedCell");
            }
        }
    }
})();

com.yscjp.jsapp.JSTable = function (id, initFrom, arg0, arg1) {

    this.initJSTableBase(id, initFrom, arg0, arg1);
};
com.yscjp.jsapp.JSTable.prototype = new com.yscjp.jsapp.JSTableBase();

//com.yscjp.jsapp.util.StackTrace.insert(com.yscjp.jsapp.JSTableBase, "com.yscjp.jsapp.JSTableBase");
/*com.yscjp.jsapp.util.StackTrace.insert(com.yscjp.jsapp.JSTable, "com.yscjp.jsapp.JSTable");*/
