package com.yscjp.webtools.gateway.servlet;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpCookie;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicHeader;

import com.yscjp.webtools.share.servlet.AutoregisterServletBase;


public class GatewayServlet extends AutoregisterServletBase {

    private static final long serialVersionUID = 1L;
    private static final Log LOG = LogFactory.getLog(GatewayServlet.class);

    @Override
    protected void doGet(HttpServletRequest request,
                         HttpServletResponse response) throws ServletException, IOException {

        execute(request, response, false);
    }


    @Override
    protected void doPost(HttpServletRequest request,
                          HttpServletResponse response) throws ServletException, IOException {

        execute(request, response, true);
    }


    private void execute(HttpServletRequest request,
                         HttpServletResponse response,
                         boolean isPost) throws ServletException, IOException {

        String fullRequestURL = getFullRequestURL(request);

        LOG.info(fullRequestURL);
        LOG.debug("PathTranslated:" + request.getPathTranslated());
        LOG.debug("RequestURI:" + request.getRequestURI());
        LOG.debug("Scheme:" + request.getScheme());
        LOG.debug("ServletPath:" + request.getServletPath());
        LOG.debug("ContextPath:" + request.getContextPath());
        LOG.debug("PathInfo:" + request.getPathInfo());
        LOG.debug("QueryString:" + request.getQueryString());

        String requestTarget = getRequestTarget(request);

        LOG.debug("RequestTarget:" + requestTarget);

        HttpUriRequest httpMethod = null;

        if (isPost) {

            HttpPost httpPost = new HttpPost(requestTarget);
            InputStream clientInput = request.getInputStream();
            InputStreamEntity requestEntity = new InputStreamEntity(clientInput);
            httpPost.setEntity(requestEntity);
            httpMethod = httpPost;

        } else {

            HttpGet httpGet = new HttpGet(requestTarget);
            httpMethod = httpGet;
        }

//        HttpGet httpGet = new HttpGet(getRequestTarget(request));
        HttpClient httpClient = createHttp(request);
        HttpResponse serverResponse = httpClient.execute(httpMethod);
        StatusLine statusLine = serverResponse.getStatusLine();
        int statusCode = statusLine.getStatusCode();
        response.setStatus(statusCode);

        Header[] headers = serverResponse.getAllHeaders();
        for (Header header : headers) {

            String name = header.getName();
            String value = header.getValue();

            LOG.debug("< Response Header:" + name + "/" + value);

            if ("location".equalsIgnoreCase(name)) {

                value = getContext(request) + value;
                LOG.debug("Location:" + value);

            } else if ("set-cookie".equalsIgnoreCase(name)) {

                URL targetURL = new URL(requestTarget);

                value = parseCookie(value, targetURL.getProtocol() + "://" + targetURL.getAuthority());
                if (null == value) {

                    continue;
                }
            }

            response.setHeader(name, value);
        }

        HttpEntity entity = serverResponse.getEntity();
        if (null != entity) {

            byte[] buffer = new byte[1024];
            InputStream in = entity.getContent();
            OutputStream out = response.getOutputStream();
            while (true) {

                int readLength = in.read(buffer);
                if (-1 == readLength) {
                    break;
                }
                out.write(buffer, 0, readLength);
            }
            in.close();
            out.close();
        }
    }


    @Override
    public String getRegistrationName() {

        return "Gateway";
    }


    @Override
    public String getRegistrationURL() {

        return "/*";
    }


    private String parseCookie(String value, String targetHost) {

        List<HttpCookie> httpCookies = HttpCookie.parse(value);
        if (0 == httpCookies.size()) {

            return null;
        }

        HttpCookie httpCookie = httpCookies.get(0);
        String path = httpCookie.getPath();
        path = "/__gateway__/" + targetHost + path;
        httpCookie.setPath(path);

        String parsedValue =
                httpCookie.getName() + "=" + httpCookie.getValue() + "; " +
                "Path=" + httpCookie.getPath() + "; " +
                "HttpOnly";

        LOG.debug("Parsed Cookie Value:" + parsedValue);

        return parsedValue;
    }


    /**
     * <pre>
     * ターゲットサーブレットへのリクエスト発行.
     * </pre>
     * @param clientRequest サーブレットリクエスト
     * @return レスポンスコード
     */
    private HttpClient createHttp(HttpServletRequest clientRequest) throws IOException {

        RequestConfig requestConfig = RequestConfig.custom().setRedirectsEnabled(false).build();

        List<BasicHeader> clientHeaders = getRequestHeaders(clientRequest);
        for (BasicHeader basicHeader : clientHeaders) {

            LOG.debug("> Request Header:" + basicHeader);
        }

        Cookie[] cookies = clientRequest.getCookies();
        if (null != cookies) {

            for (Cookie cookie : cookies) {

                LOG.debug("> Cookie:" +
                          cookie.getName() + "=" +
                          cookie.getValue());
                BasicHeader basicHeader = new BasicHeader("Cookie", cookie.getName() + "=" + cookie.getValue());
                clientHeaders.add(basicHeader);
            }
        }

        HttpClient httpClient = HttpClientBuilder.create()
                .setDefaultRequestConfig(requestConfig)
                .setDefaultHeaders(clientHeaders)
                .build();

        return httpClient;
    }


    /**
     * クライアントからのリクエストヘッダを取得し、HttpClientが理解できる形式に変換して返します。
     * この中には、hostとcookieは含まれません。
     * @param clientRequest サーブレットリクエスト
     * @return HttpClientが理解できる形式のヘッダを格納したList
     */
    private List<BasicHeader> getRequestHeaders(HttpServletRequest clientRequest) throws IOException {

        List<BasicHeader> clientHeaders = new ArrayList<BasicHeader>();
        Enumeration<String> names = clientRequest.getHeaderNames();

        while (names.hasMoreElements()) {

            String name = names.nextElement();
            if ("host".equalsIgnoreCase(name) || "cookie".equalsIgnoreCase(name) || "content-length".equalsIgnoreCase(name)) {

                continue;
            }

            Enumeration<String> values = clientRequest.getHeaders(name);
            while (values.hasMoreElements()) {

                String value = values.nextElement();
                BasicHeader basicHeader = new BasicHeader(name, value);
                clientHeaders.add(basicHeader);
            }
        }

        clientHeaders.add(getHostHeader(clientRequest));

        return clientHeaders;
    }


    private BasicHeader getHostHeader(HttpServletRequest clientRequest) throws IOException {

        String requestTarget = getRequestTarget(clientRequest);
        URL url = new URL(requestTarget);
        String host = url.getHost();
        int port = url.getPort();

        if (-1 != port) {

            host = host + ":" + port;
        }

        BasicHeader basicHeader = new BasicHeader("host", host);

        return basicHeader;
    }


    private String getRequestTarget(HttpServletRequest clientRequest) {

        String requestURI = clientRequest.getRequestURI();
        String queryString = clientRequest.getQueryString();
        if (null != queryString) {

            requestURI = requestURI + "?" + queryString;
        }
        String contextPath = clientRequest.getContextPath();
        String requestTarget = requestURI.replace(contextPath + "/", "");

        return requestTarget;
    }


    private String getContext(HttpServletRequest clientRequest) {

        String requestURL = clientRequest.getRequestURL().toString();
        String requestURI = clientRequest.getRequestURI();
        int endIndex = requestURL.indexOf(requestURI);
        String context = requestURL.substring(0, endIndex);
        String contextPath = clientRequest.getContextPath();

        LOG.info("requestURL:" + requestURL);
        LOG.info("pathInfo:" + requestURI);
        LOG.info("endIndex:" + endIndex);

        return context + contextPath + "/";
    }
}
